import random
import numpy as np
import copy

import torch.utils.data as Data
import networkx as nx
from networkx import from_numpy_array
import torch
from sklearn.metrics.pairwise import cosine_similarity
import math


def make_data_contrastive_square(df_expr, neg_rate=0.05, pos_rate=0.05):
    # obtain pos and expr info
    spot_pos_all = df_expr.iloc[:, 1:3].to_numpy()
    spot_info_all = df_expr.iloc[:, 3:].to_numpy()

    # calculate cosine matrix
    expr = copy.deepcopy(spot_info_all)
    similarity_matrix = cosine_similarity(expr)
    hist, bin_edges = np.histogram(similarity_matrix, bins=20, range=(0.0, 1.0))

    near_similarity_all = []
    indices_1_all = []
    for i, spot_pos_i in enumerate(spot_pos_all):
        indices_1 = np.where((np.abs(spot_pos_all[:, 0] - spot_pos_i[0]) <= 1) & (np.abs(spot_pos_all[:, 1] - spot_pos_i[1]) <= 1))[0]
        indices_1_all.append(indices_1)
        near_similarity_all += list(similarity_matrix[i, indices_1])
    sorted_lst = np.sort(near_similarity_all)[::-1]
    sorted_lst = sorted_lst[len(spot_pos_all):]
    top_thres = sorted_lst[int(len(sorted_lst) * pos_rate)]

    similarity_all = similarity_matrix.flatten()
    sigma = len(spot_pos_all) * neg_rate
    low_thres = np.sort(similarity_all)[int(len(spot_pos_all) * sigma)]

    # find positive and negative samples
    train_data = []
    pos_statistics = []
    neg_statistics = []

    if sigma > 200:
        neg_num = 200
    else:
        neg_num = int(sigma)
    for i in range(len(spot_pos_all)):
        # find positive samples
        indices_2 = np.where(similarity_matrix[i, :] > top_thres)[0]  # high expression similarity
        indices = np.intersect1d(indices_1_all[i], indices_2)
        pos_statistics.append(len(indices))
        positive_spot = np.array(spot_info_all)[indices, :]

        # find negative samples
        indices = np.where(similarity_matrix[i, :] <= low_thres)[0]
        if len(indices) >= 1 and len(indices) <= neg_num:
            neg_statistics.append(len(indices))
        elif len(indices) < 1:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:1]
            neg_statistics.append(len(indices))
        else:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:neg_num]
            neg_statistics.append(len(indices))
        negative_spot = np.array(spot_info_all)[indices, :]

        # save info
        train_data.append([spot_info_all[i], positive_spot, negative_spot])

    return train_data


def make_data_contrastive_hexagon(df_expr, neg_rate=0.05, pos_rate=0.05):
    # obtain pos and expr info
    spot_pos_all = df_expr.iloc[:, 1:3].to_numpy()
    spot_info_all = df_expr.iloc[:, 3:].to_numpy()

    # calculate cosine matrix
    expr = copy.deepcopy(np.array(spot_info_all))
    similarity_matrix = cosine_similarity(expr)
    hist, bin_edges = np.histogram(similarity_matrix, bins=20, range=(0.0, 1.0))

    near_similarity_all = []
    indices_1_all = []
    for i, spot_pos_i in enumerate(spot_pos_all):
        indices_1 = np.where((np.abs(spot_pos_all[:, 0] - spot_pos_i[0]) <= 1) & (np.abs(spot_pos_all[:, 1] - spot_pos_i[1]) <= 2))[0]
        indices_1_all.append(indices_1)
        near_similarity_all += list(similarity_matrix[i, indices_1])
    sorted_lst = np.sort(near_similarity_all)[::-1]
    sorted_lst = sorted_lst[len(spot_pos_all):]
    top_thres = sorted_lst[int(len(sorted_lst) * pos_rate)]

    similarity_all = similarity_matrix.flatten()
    sigma = len(spot_pos_all) * neg_rate
    low_thres = np.sort(similarity_all)[int(len(spot_pos_all) * sigma)]

    # find positive and negative samples
    train_data = []
    pos_statistics = []
    neg_statistics = []

    if sigma > 200:
        neg_num = 200
    else:
        neg_num = int(sigma)
    for i in range(len(spot_pos_all)):
        # find positive samples
        indices_2 = np.where(similarity_matrix[i, :] > top_thres)[0]  # high expression similarity
        indices = np.intersect1d(indices_1_all[i], indices_2)
        pos_statistics.append(len(indices))
        positive_spot = np.array(spot_info_all)[indices, :]

        # find negative samples
        indices = np.where(similarity_matrix[i, :] <= low_thres)[0]
        if len(indices) >= 1 and len(indices) <= neg_num:
            neg_statistics.append(len(indices))
        elif len(indices) < 1:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:1]
            neg_statistics.append(len(indices))
        else:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:neg_num]
            neg_statistics.append(len(indices))
        negative_spot = np.array(spot_info_all)[indices, :]

        # save info
        train_data.append([spot_info_all[i], positive_spot, negative_spot])

    return train_data


def make_data_contrastive_others(df_expr, neg_rate=0.05, pos_rate=0.05):
    # obtain pos and expr info
    spot_pos_all = df_expr.iloc[:, 1:3].to_numpy()
    spot_info_all = df_expr.iloc[:, 3:].to_numpy()

    # calculate cosine matrix
    expr = copy.deepcopy(spot_info_all)
    similarity_matrix = cosine_similarity(expr)
    hist, bin_edges = np.histogram(similarity_matrix, bins=20, range=(0.0, 1.0))

    distance = np.sum((spot_pos_all - spot_pos_all[:, np.newaxis]) ** 2, axis=2)
    indices_1_all = np.argpartition(distance, 11)[:, :11]
    near_similarity_all = similarity_matrix[np.arange(len(spot_pos_all))[:, np.newaxis], indices_1_all].flatten()
    sorted_lst = np.sort(near_similarity_all)[::-1]
    sorted_lst = sorted_lst[len(spot_pos_all):]
    top_thres = sorted_lst[int(len(sorted_lst) * pos_rate)]

    similarity_all = similarity_matrix.flatten()
    sigma = len(spot_pos_all) * neg_rate
    low_thres = np.sort(similarity_all)[int(len(spot_pos_all) * sigma)]

    # find positive and negative samples
    train_data = []
    pos_statistics = []
    neg_statistics = []

    if sigma > 200:
        neg_num = 200
    else:
        neg_num = int(sigma)

    for i in range(len(spot_pos_all)):
        # find positive samples
        indices_2 = np.where(similarity_matrix[i, :] > top_thres)[0]  # high expression similarity
        indices = np.intersect1d(indices_1_all[i], indices_2)
        pos_statistics.append(len(indices))
        positive_spot = spot_info_all[indices, :]

        # find negative samples
        indices = np.where(similarity_matrix[i, :] <= low_thres)[0]
        if len(indices) >= 1 and len(indices) <= neg_num:
            neg_statistics.append(len(indices))
        elif len(indices) < 1:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:1]
            neg_statistics.append(len(indices))
        else:
            indices = np.argsort(similarity_matrix[i, :], axis=None)[0:neg_num]
            neg_statistics.append(len(indices))
        negative_spot = spot_info_all[indices, :]

        # save info
        train_data.append([spot_info_all[i], positive_spot, negative_spot])

    return train_data


def make_test_data_contrastive(df_expr):
    spot_pos_all = []
    spot_info_all = []

    # obtain pos and expr info
    for i in range(len(df_expr)):
        spot_pos_all.append(df_expr.iloc[i, 1:3].to_numpy().astype(float))
        spot_info_all.append(df_expr.iloc[i, 3:].to_numpy().astype(float))

    test_data = spot_info_all

    return test_data, spot_pos_all


class train_Dataset_contrastive(Data.Dataset):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, index):
        spot = copy.deepcopy(self.data[index][0])
        positive_sample = copy.deepcopy(self.data[index][1])
        negative_sample = copy.deepcopy(self.data[index][2])
        spot = torch.from_numpy(spot).float().unsqueeze(0)
        positive_sample = torch.from_numpy(positive_sample).float()
        negative_sample = torch.from_numpy(negative_sample).float()
        # patch_noise = add_noise(patch)

        return spot, positive_sample, negative_sample, len(positive_sample), len(negative_sample)

    def __len__(self):
        return len(self.data)


class test_Dataset_contrastive(Data.Dataset):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, index):
        spot = copy.deepcopy(self.data[index]).astype(np.float32)
        spot = torch.from_numpy(spot)

        return spot

    def __len__(self):
        return len(self.data)






import numpy as np
import pandas as pd
from scipy.stats import pearsonr


def cell_type_matching(df_sc, df_beta):
    gt_cell_type_num = df_sc.shape[0]
    pred_cell_type_num = df_beta.shape[0]
    cell_type_gt = df_sc.index.tolist()

    # calculate correlation between gt and pred
    overlapping_genes = list(set(df_sc.columns.tolist()) & set(df_beta.columns.tolist()))
    cell_type_correlation = np.zeros([gt_cell_type_num, pred_cell_type_num])
    df_sc = df_sc.loc[:, overlapping_genes]
    df_beta = df_beta.loc[:, overlapping_genes]
    df_sc /= (df_sc.to_numpy().sum(axis=1, keepdims=True))
    df_beta /= (df_beta.to_numpy().sum(axis=1, keepdims=True))
    df_sc /= (df_sc.sum(axis=0) + 1e-6)
    df_beta /= (df_beta.sum(axis=0) + 1e-6)
    for i in range(gt_cell_type_num):
        sc_expr = df_sc.iloc[i, :].to_numpy()
        for j in range(pred_cell_type_num):
            pred_expr = df_beta.iloc[j, :].to_numpy()
            cell_type_correlation[i, j] = pearsonr(sc_expr, pred_expr)[0]
    df_corr = pd.DataFrame(cell_type_correlation, index=cell_type_gt, columns=['X' + str(k) for k in range(pred_cell_type_num)])

    # match gt and pred
    cell_type_score_copy = cell_type_correlation.copy()
    df_matching_results = pd.DataFrame(index=cell_type_gt, columns=['cell_type_index', 'corr'])
    for i in range(gt_cell_type_num):
        max_value = np.max(cell_type_score_copy)
        if max_value == -100:
            break
        max_indices = np.transpose(np.where(cell_type_score_copy == max_value))
        for row_index, col_index in max_indices:
            df_matching_results.iloc[row_index, 0] = col_index
            df_matching_results.iloc[row_index, 1] = max_value
            cell_type_score_copy[row_index] = -100
            cell_type_score_copy[:, col_index] = -100

    return df_matching_results, df_corr


import pandas as pd


def calc_cell_type_expr(df_sc, save_dir):
    data_all = []
    cell_type = df_sc.iloc[:, 0].unique()
    t = df_sc.iloc[:, 1:].to_numpy()
    df_sc.iloc[:, 1:] = t / t.sum(axis=1, keepdims=True)
    for cell_type_i in cell_type:
        expr_i = df_sc.loc[df_sc['cell_type'] == cell_type_i].iloc[:, 1:].mean(axis=0)
        data_all.append(expr_i)
    df = pd.concat(data_all, axis=1).T
    df.index = cell_type
    df.to_csv(save_dir + 'sc_based_cell_type_expression.csv')

    return df


import pandas as pd
import torch
import math
import numpy as np
import matplotlib.pyplot as plt


def replace_top_k(column):
    sorted_column = column.sort_values(ascending=False).reset_index(drop=True)
    if len(sorted_column) > 10:
        eleventh_value = sorted_column.iloc[10]  
        top_10_indices = column.nlargest(10).index
        column.loc[top_10_indices] = eleventh_value
    return column


def calculate_metagenes(df_data, df_beta, topk1=5, topk2=10):
    df_marker = pd.DataFrame()
    df_logFC = pd.DataFrame()
    top_k1_marker_all = []
    for i in range(len(df_beta)):
        num = i
        df_beta_i = df_beta.iloc[i, :]
        sel = list(set(range(df_beta.shape[0])) - set([i]))
        df_sel = df_beta.iloc[sel, :]
        fc = df_beta.iloc[num, :] / df_sel.mean(axis=0)
        log2fc = fc.apply(lambda x: math.log2(x))
        sort = log2fc.sort_values(ascending=False)
        df_marker[f'{i}'] = sort.iloc[:topk2].index
        df_logFC[f'{i}'] = log2fc
        top_k1_marker_all += sort.iloc[:topk1].index.tolist()

    df_metagene = pd.DataFrame(np.zeros([len(df_data), len(df_beta)]), columns=[f'{i}' for i in range(len(df_beta))])
    df_expr = df_data.iloc[:, 3:]
    df_expr = df_expr.apply(replace_top_k)
    df_expr = df_expr.apply(lambda col: col / col.max() if col.max() != 0 else col)
    for i in range(len(df_beta)):
        marker_i_list = df_marker.iloc[:, i].tolist()
        df_metagene[f'{i}'] = df_expr.loc[:, marker_i_list].mean(axis=1)
        df_metagene[f'{i}'] = df_metagene[f'{i}'] / df_metagene[f'{i}'].max()
        
    return df_logFC, top_k1_marker_all, df_metagene 


def metagene_plot(df_logFC, df_beta, top_k1_marker_all, sel_celltype, topk1=5):
    top_marker_all = []
    
    for celltype in sel_celltype:
        start_idx = celltype * topk1
        end_idx = (celltype + 1) * topk1

        celltype_data = top_k1_marker_all[start_idx:end_idx]
        top_marker_all.extend(celltype_data)
        
    df_top_FC = df_logFC.loc[top_marker_all]
    df_top_FC = df_top_FC.iloc[:, sel_celltype]
    df_top_beta = df_beta.T.loc[top_marker_all, sel_celltype]
    FC_arr = df_top_FC.T.values.flatten()
    beta_arr = df_top_beta.T.values.flatten()
    beta_arr /= beta_arr.max()
    x_label = []
    y_label = []
    for i in range(len(df_top_FC.columns.tolist())):
        for j in range(len(df_top_FC.index.tolist())):
            x_label.append(f'X{sel_celltype[i]}')
            y_label.append(df_top_FC.index.tolist()[j])

    fig, ax = plt.subplots(figsize=(4, 5))
    plt.subplots_adjust(left=0.25, right=0.9, top=0.98, bottom=0.15)
    im = ax.scatter(x_label, y_label, s=FC_arr * 30 * 2, c=beta_arr, marker='.', cmap='plasma')
    im.set_clim(0, 1)
    plt.xticks(fontsize=10, fontname='Arial')
    plt.yticks(fontsize=10, fontstyle='italic', fontname='Arial')
    cbar = plt.colorbar(im, ax=ax, cmap='plasma', location='top', pad=0.01)
    cbar.set_label('Deconvolved Expressions', labelpad=5, fontsize=10, fontfamily='Arial')
    cbar.ax.tick_params(labelsize=7)
    legend_sizes = [40, 80, 120]
    legend_labels = ['2', '4', '6']
    handles = [plt.scatter([], [], s=size, edgecolor='black', color='gray') for size in legend_sizes]
    legend = plt.legend(handles, legend_labels, title="Log$_{2}$fold_change", title_fontsize=10, frameon=False,
                        loc="lower center", bbox_to_anchor=(0.5, -0.23), ncol=3, prop={'family': 'Arial', 'size': 10})
    plt.show()